import bisect
import json
import os
import stat
from fnmatch import fnmatch

from kamikaze3.common.filesystem import APP_USER_UID
from kamikaze3.filesystem_scan.metadata import metadata_for_dir, metadata_for_file
from kamikaze3.filesystem_scan.settings import (
    BIGGEST_DIRS_LIMIT,
    BIGGEST_FILES_LIMIT,
    FS_SCAN_OUTPUT_FILE,
    IGNORE_PATTERNS,
    TARGET_DIRS,
)


def should_ignore(path, patterns, file_stat=None):
    """
    Ignore the files that match any of the patterns,
    and if file_stat is provided, ignore all files
    that do not belong to the app user.
    """
    if file_stat:
        if file_stat.st_uid != APP_USER_UID:
            return True
    for pattern in patterns:
        if fnmatch(path, pattern):
            return True
    return False


def add_to_sorted_list(elem, sorted_list, truncate_len=0, key_name="size"):
    """
    Inserts an element into a list while maintaining sorted order, and optionally truncates the list.

    Args:
        elem (dict): The element to insert into the sorted list.
        sorted_list (list of dict): The list to insert into. Must be sorted by `key_name`.
        truncate_len (int, optional): If greater than 0, keeps only the last `truncate_len` elements after insertion. Defaults to 0 (no truncation).
        key_name (str, optional): The dictionary key to use for sorting. Defaults to "size".

    Returns:
        None: The input list is modified in place.
    """
    bisect.insort(sorted_list, elem, key=lambda x: x[key_name])
    sorted_list[:] = sorted_list[-truncate_len:]


def add_to_list_if_big_dir(dirpath, dir_cardinality, biggest_dirs, limit=0):
    """
    Adds directory metadata to a list of the largest directories if it qualifies.

    The directory is added only if there's room in the list (based on limit) or if its
    cardinality (number of children) is greater than the smallest cardinality currently
    in the list. The list is kept sorted by the "children" count and optionally truncated
    to retain only the top N directories.

    Args:
        dirpath (str): Path to the directory being considered.
        dir_cardinality (int): Number of child elements in the directory.
        biggest_dirs (list of dict): List of metadata dictionaries for the largest directories.
        limit (int): Number to truncate the list to, 0 for no truncation

    Returns:
        None: The `biggest_dirs` list is modified in place.
    """
    list_not_full = limit == 0 or len(biggest_dirs) < limit
    if list_not_full or dir_cardinality > biggest_dirs[0]["children"]:
        add_to_sorted_list(
            metadata_for_dir(dirpath, dir_cardinality),
            biggest_dirs,
            truncate_len=limit,
            key_name="children",
        )


def add_to_list_if_broken_symlink(file_path, file_stat, broken_symlinks):
    """
    Checks if a file is a broken symbolic link and adds it to a list if so.

    Args:
        file_path (str): Path to the file to check.
        file_stat (os.stat_result): Result of a `lstat()` call on the file.
        broken_symlinks (list of str): List to which broken symlink paths are appended.

    Returns:
        bool: True if the file is a broken symlink and was added to the list, False otherwise.
    """
    is_symlink = stat.S_ISLNK(file_stat.st_mode)
    is_broken = not os.path.exists(file_path)
    if is_symlink and is_broken:
        broken_symlinks.append(file_path)
        return True
    return False


def add_to_list_if_big_file(file_path, file_stat, biggest_files, limit=0):
    """
    Adds file metadata to a list of the largest files if it qualifies.

    The file metadata is added only if there's room in the list (based on limit) or if its size
    is greater than the smallest size currently in the list. The list is maintained in sorted
    order by file size and optionally truncated to retain only the top N largest files.

    Args:
        file_path (str): Path to the file being considered.
        file_stat (os.stat_result): Result of a `lstat()` call on the file.
        biggest_files (list of dict): List of metadata dictionaries for the largest files.
        limit (int): Number to truncate the list to, 0 for no truncation

    Returns:
        None: The `biggest_files` list is modified in place.
    """
    list_not_full = limit == 0 or len(biggest_files) < limit
    file_size = file_stat.st_size
    if list_not_full or file_size > biggest_files[0]["size"]:
        add_to_sorted_list(
            metadata_for_file(file_path, file_stat),
            biggest_files,
            truncate_len=limit,
            key_name="size",
        )


def scan(root, biggest_dirs, biggest_files, broken_symlinks):
    for dirpath, dirs, files in os.walk(root, topdown=True, followlinks=False):
        # modify dirs in place to prevent os.walk from recursing into ignored directories
        dirs[:] = [
            d
            for d in dirs
            if not should_ignore(os.path.join(dirpath, d), IGNORE_PATTERNS)
        ]

        add_to_list_if_big_dir(
            dirpath, len(files) + len(dirs), biggest_dirs, limit=BIGGEST_DIRS_LIMIT
        )

        for filename in files:
            file_path = os.path.join(dirpath, filename)
            file_stat = os.lstat(file_path)
            if should_ignore(file_path, IGNORE_PATTERNS, file_stat):
                continue
            if add_to_list_if_broken_symlink(file_path, file_stat, broken_symlinks):
                continue
            add_to_list_if_big_file(
                file_path, file_stat, biggest_files, limit=BIGGEST_FILES_LIMIT
            )


def run_scan():
    biggest_files = []
    broken_symlinks = []
    biggest_dirs = []

    for target_dir in TARGET_DIRS:
        scan(target_dir, biggest_dirs, biggest_files, broken_symlinks)

    scan_results = {
        "big_dirs": biggest_dirs,
        "big_files": biggest_files,
        "broken_symlinks": broken_symlinks,
    }
    with open(FS_SCAN_OUTPUT_FILE, "w") as file:
        json.dump(scan_results, file)


if __name__ == "__main__":
    run_scan()
