import json
import logging
import os
from os.path import join
import shutil
import subprocess
import datetime

from kamikaze3.importer.abstract import ImporterSubClass

log = logging.getLogger(__name__)


class DeploymentSetup(ImporterSubClass):
    """
    Set up a Hypernode deployment structure in /data/web/apps/[domain]
    This creates a capistrano-like structure with shared and releases directories
    specifically for Magento 2 installations.
    """

    SHARED_FILES = [
        "app/etc/env.php",
        "pub/media",
        "pub/errors/local.xml",
        "pub/.user.ini",
        "var/log",
        "var/report",
        "var/session",
    ]

    def __init__(self, importer):
        """
        Initialize the DeploymentSetup.
        :param obj importer: The Importer object where we keep the global state.
        :return None:
        """
        super().__init__(importer=importer)

        if self.importer.site_type != "magento2":
            raise ValueError(
                "DeploymentSetup is only supported for Magento 2 installations"
            )

        self.deployed = False
        self.domain = self.determine_domain()
        self.app_root = "/data/web/"
        self.apps_dir = join(self.app_root, "apps")
        self.site_dir = join(self.apps_dir, self.domain)
        self.shared_dir = join(self.site_dir, "shared")
        self.releases_dir = join(self.site_dir, "releases")
        self.current_release_dir = join(self.releases_dir, "1")
        self.current_symlink = join(self.site_dir, "current")
        self.public_symlink = join(self.app_root, "public")
        self.source_dir = self.importer.remote_path

    def determine_domain(self):
        """
        Determine the domain name to use for the deployment folder.
        Uses --deploy-domain if specified, otherwise extracts from /etc/hypernode/app.json
        :return str domain: Domain name
        """
        if self.importer.deploy_domain:
            return self.importer.deploy_domain

        try:
            with open("/etc/hypernode/app.json") as app_json:
                app_data = json.load(app_json)
                return app_data.get("hn_fqdn", "default")
        except (OSError, ValueError):
            log.warning(
                "Could not read /etc/hypernode/app.json, using 'default' as domain name"
            )
            return "default"

    def create_directory_structure(self):
        """
        Create the Magento 2 deployment directory structure
        :return None:
        """
        log.info("Creating Magento 2 deployment directory structure")

        os.makedirs(self.apps_dir, exist_ok=True)
        os.makedirs(self.site_dir, exist_ok=True)
        os.makedirs(self.releases_dir, exist_ok=True)
        os.makedirs(self.current_release_dir, exist_ok=True)

        for path in self.SHARED_FILES:
            if path.endswith("/"):
                dir_path = path
            else:
                dir_path = os.path.dirname(path)
            os.makedirs(join(self.shared_dir, dir_path), exist_ok=True)

    def move_files_to_shared(self):
        """
        Move specific Magento 2 files/directories from the source to the shared directory
        :return None:
        """
        log.info("Moving Magento 2 files to shared directory")

        shared_paths = []
        for path in self.SHARED_FILES:
            shared_paths.append({"src": path, "dst": join(self.shared_dir, path)})

        for path in shared_paths:
            src_path = join(self.source_dir, path["src"])
            log.debug(f"Moving {src_path} to {path['dst']}")
            try:
                os.makedirs(os.path.dirname(path["dst"]), exist_ok=True)
                shutil.move(src_path, path["dst"])
            except (OSError, shutil.Error) as e:
                raise RuntimeError(
                    f"Failed to move {src_path} to {path['dst']}: {str(e)}"
                )

    def move_files_to_release(self):
        """
        Move all remaining Magento 2 files from the source to the release directory
        :return None:
        """
        log.info(
            f"Moving Magento 2 files to release directory: {self.current_release_dir}"
        )

        rsync_args = ["/usr/bin/rsync", "-a", "--delete"]

        for path in self.SHARED_FILES:
            rsync_args.append(f"--exclude=/{path}")

        rsync_args.extend(
            [
                f"{self.source_dir}/",
                f"{self.current_release_dir}/",
            ]
        )

        try:
            subprocess.check_call(rsync_args)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f"Failed to move Magento 2 files to release directory: {str(e)}"
            )

    def create_symlinks(self):
        """
        Create symlinks from the release directory to the shared directory for Magento 2
        :return None:
        """
        log.info("Creating Magento 2 symlinks")

        symlinks = []
        for path in self.SHARED_FILES:
            symlinks.append(
                {
                    "target": join(self.shared_dir, path),
                    "link_name": join(self.current_release_dir, path),
                }
            )

        for link in symlinks:
            os.makedirs(os.path.dirname(link["link_name"]), exist_ok=True)

            if os.path.exists(link["link_name"]) or os.path.islink(link["link_name"]):
                if os.path.isdir(link["link_name"]) and not os.path.islink(
                    link["link_name"]
                ):
                    shutil.rmtree(link["link_name"])
                else:
                    os.remove(link["link_name"])

            os.symlink(link["target"], link["link_name"])
            log.debug(
                f"Created Magento 2 symlink: {link['link_name']} -> {link['target']}"
            )

    def create_deployment_symlinks(self):
        """
        Create additional deployment symlinks:
        1. current -> releases/1 (in the deployment directory)
        2. /data/web/public -> /data/web/apps/[domain]/releases/1/pub
        :return None:
        """
        log.info("Creating deployment symlinks")

        if os.path.exists(self.current_symlink) or os.path.islink(self.current_symlink):
            if os.path.isdir(self.current_symlink) and not os.path.islink(
                self.current_symlink
            ):
                shutil.rmtree(self.current_symlink)
            else:
                os.remove(self.current_symlink)

        os.symlink(self.current_release_dir, self.current_symlink)
        log.debug(
            f"Created 'current' symlink: {self.current_symlink} -> {self.current_release_dir}"
        )

        public_target = join(self.current_release_dir, "pub")
        if os.path.exists(self.public_symlink) or os.path.islink(self.public_symlink):
            if os.path.isdir(self.public_symlink) and not os.path.islink(
                self.public_symlink
            ):
                shutil.rmtree(self.public_symlink)
            else:
                os.remove(self.public_symlink)

        os.symlink(public_target, self.public_symlink)
        log.debug(f"Created 'public' symlink: {self.public_symlink} -> {public_target}")

    def create_dep_directory_and_files(self):
        """
        Create the .dep directory and required files:
        1. apps/[domain]/.dep/latest_release with content "1"
        2. apps/[domain]/.dep/releases_log with a JSON entry for the current date
        :return None:
        """
        log.info("Creating .dep directory and files")

        dep_dir = join(self.site_dir, ".dep")
        try:
            os.makedirs(dep_dir, exist_ok=True)
        except OSError as e:
            raise RuntimeError(
                f"Failed to create .dep directory at {dep_dir}: {str(e)}"
            )

        latest_release_path = join(dep_dir, "latest_release")
        try:
            with open(latest_release_path, "w") as f:
                f.write("1")
            log.debug(f"Created latest_release file: {latest_release_path}")
        except (OSError, IOError) as e:
            raise RuntimeError(
                f"Failed to create latest_release file at {latest_release_path}: {str(e)}"
            )

        releases_log_path = join(dep_dir, "releases_log")
        current_date = datetime.datetime.now().isoformat()
        release_entry = {
            "created_at": current_date,
            "release_name": "1",
            "user": "ci",
            "target": "HEAD",
        }

        try:
            with open(releases_log_path, "w") as f:
                f.write(json.dumps(release_entry))
            log.debug(f"Created releases_log file: {releases_log_path}")
        except (OSError, IOError) as e:
            raise RuntimeError(
                f"Failed to create releases_log file at {releases_log_path}: {str(e)}"
            )

    def setup_deployment(self):
        """
        Set up the Magento 2 deployment structure.
        :return None:
        """
        if not self.importer.hypernode_deploy:
            log.debug("Hypernode deployment not requested, skipping")
            return

        if self.importer.site_type != "magento2":
            log.warning(
                "Deployment setup is only supported for Magento 2 installations, skipping"
            )
            return

        log.info(f"Setting up Magento 2 deployment for domain: {self.domain}")

        self.create_directory_structure()
        self.move_files_to_shared()
        self.move_files_to_release()
        self.create_symlinks()
        self.create_deployment_symlinks()
        self.create_dep_directory_and_files()

        log.info(f"Magento 2 deployment setup complete at: {self.site_dir}")
        self.deployed = True
