import logging
import os
from datetime import datetime
from pathlib import Path
from typing import Optional, Tuple, Dict

logger = logging.getLogger(__name__)

DEFAULT_PHP_VERSION = "7.4"
JOB_POST_WAIT = 10

SERVICE_VERSION_MAP: Dict[str, Dict[str, str]] = {
    "elasticsearch": {
        "5": "5.2",
        "6": "6.x",
        "7": "7.x",
    },
    "redis": {
        "5": "5.0",
        "6": "6.x",
        "7": "7.x",
    },
    "rabbitmq": {
        "3": "3.10",
    },
    "mysql": {
        "5": "5.7",
        "8": "8.0",
    },
}


def extract_version_components(version_str: str) -> Tuple[Optional[str], Optional[str]]:
    """
    Extract major and minor version components from a version string.

    :param str version_str: Version string (e.g., "5.7.32", "6.2", "7")
    :return tuple: (major_version, minor_version) both as strings
    """
    if not version_str:
        return None, None

    parts = version_str.split(".")
    major_version = parts[0]
    minor_version = parts[1] if len(parts) > 1 else "0"

    return major_version, minor_version


def standardize_version(version_str: str, service_type: str) -> Optional[str]:
    """
    Standardize version string format based on service type and major version.

    :param str version_str: Raw version string (e.g. "5.0.1", "6.2", "7")
    :param str service_type: Type of service ("redis", "elasticsearch", "rabbitmq", "mysql")
    :return str: Standardized version string
    """
    if not version_str:
        return None

    major_version, _ = extract_version_components(version_str)
    return SERVICE_VERSION_MAP.get(service_type, {}).get(
        major_version, f"{major_version}.0"
    )


def _read_remote_file(importer, file_path: str) -> Optional[str]:
    """
    Read a remote file and return its content, or None if the file can't be read
    :param importer: Importer object to connect to remote node
    :param file_path: Path to the file to read
    :return: File content as string, or None if the file can't be read
    """
    return_code, stdout, _ = importer.ssh_conn.driver.exec_command(f"cat {file_path}")
    if return_code != 0 or not stdout:
        return None
    return stdout.decode("utf-8")


def is_hipex_node(importer) -> bool:
    """
    Check whether the node resembles signs of a Hipex node
    :param importer: Importer object to connect to remote node
    to the node
    :rtype: bool
    """
    return_code, stdout, _ = importer.ssh_conn.driver.exec_command("hostname")
    if return_code != 0:
        logger.warning("Can't parse hostname. Assuming not a Hipex node")
        return False
    is_hipex = stdout.strip().decode("utf-8").endswith(".hipex.io")
    if is_hipex:
        logger.info("Found remote host to be a Hipex node")
    return is_hipex


def create_hipex_state_dir(base_dir=None) -> Path:
    """
    Create a directory to store Hipex state files.

    :param base_dir: Optional base directory to use instead of the default /data/web
    :return: Path to the created directory
    """
    date_of_import = datetime.now()
    if base_dir is None:
        base_dir = Path("/data/web")
    backup_dir = base_dir / f"importer_hipex_state/{date_of_import.isoformat()}"
    logger.info(f"Preserving Hipex state in {backup_dir}")
    backup_dir.mkdir(parents=True, exist_ok=True)
    return backup_dir
