import logging
import time

from kamikaze3.log.main import block_until_last_flow_done
from kamikaze3.systemctl.settings import set_attribute
from kamikaze3.importer.hipex.common import is_hipex_node, create_hipex_state_dir
from kamikaze3.importer.hipex.php import get_hipex_php_version, backup_hipex_php_config
from kamikaze3.importer.hipex.mysql import (
    get_remote_mysql_version,
    backup_hipex_mysql_config,
)
from kamikaze3.importer.hipex.redis import backup_hipex_redis_config
from kamikaze3.importer.hipex.nginx import backup_hipex_nginx_config
from kamikaze3.importer.hipex.supervisor import backup_hipex_supervisor_config
from kamikaze3.importer.hipex.crontab import backup_hipex_crontab
from kamikaze3.importer.hipex.varnish import backup_hipex_varnish_config
from kamikaze3.importer.hipex.rabbitmq import backup_hipex_rabbitmq_config
from kamikaze3.importer.hipex.elasticsearch import backup_hipex_elasticsearch_config
from kamikaze3.importer.hipex.docker import backup_hipex_docker_config

logger = logging.getLogger(__name__)

JOB_POST_WAIT = 10


def backup_hipex_state(importer):
    """
    Create a copy of all existing state on the Hipex server for future debugging. This can
    be helpful when comparing the previous state on the Hipex server to the new state on
    the Hypernode.
    """
    backup_dir = create_hipex_state_dir()
    backup_hipex_crontab(importer=importer, backup_dir=backup_dir)
    backup_hipex_mysql_config(importer=importer, backup_dir=backup_dir)
    backup_hipex_nginx_config(importer=importer, backup_dir=backup_dir)
    backup_hipex_redis_config(importer=importer, backup_dir=backup_dir)
    backup_hipex_php_config(importer=importer, backup_dir=backup_dir)
    backup_hipex_supervisor_config(importer=importer, backup_dir=backup_dir)

    # New components that can be added as needed
    backup_hipex_varnish_config(importer=importer, backup_dir=backup_dir)
    backup_hipex_rabbitmq_config(importer=importer, backup_dir=backup_dir)
    backup_hipex_elasticsearch_config(importer=importer, backup_dir=backup_dir)
    backup_hipex_docker_config(importer=importer, backup_dir=backup_dir)


def setup_hipex_transfer(importer):
    """
    Detect all relevant settings at the Hipex server and set them at the Hypernode server.
    :param importer: Importer object to connect to remote node
    """
    remote_php_version = get_hipex_php_version(importer=importer)

    remote_mysql_version = get_remote_mysql_version(importer=importer)

    if importer.dry_run:
        logger.info("Would have set MySQL version to {}".format(remote_mysql_version))
    else:
        logger.info("Setting MySQL version to {}".format(remote_mysql_version))
        set_attribute("mysql_version", remote_mysql_version)

    if importer.dry_run:
        logger.info("Would have set PHP version to {}".format(remote_php_version))
    else:
        if remote_php_version:
            logger.info("Setting PHP version to {}".format(remote_php_version))
            set_attribute("php_version", remote_php_version)
        else:
            logger.info(
                "Not setting PHP version because we failed to detect "
                "which version on the Hipex server"
            )

    logger.info(
        "Waiting {} seconds until update jobs are posted..".format(JOB_POST_WAIT)
    )
    time.sleep(JOB_POST_WAIT)

    block_until_last_flow_done()
