import logging
from argparse import ArgumentParser
from time import sleep

from kamikaze3.common.env import check_dns, get_hn_common_var, get_public_ip
from kamikaze3.log.log import configure_logging
from kamikaze3.manage_floating_ip.manage_digital_ocean_floating_ip import (
    ManageDigitalOceanFloatingIP,
)
from kamikaze3.manage_floating_ip.manage_tbbm_floating_ip import (
    ManageTBBMFloatingIP,
    SafeRuntimeError,
)
from kamikaze3.manage_floating_ip.settings import PROVIDER_DIGITAL_OCEAN, PROVIDER_TBBM
from kamikaze3.utils import TimeoutException, wait_for

logger = logging.getLogger(__name__)


def verify_outbound_ip_is_expected(expected_ip):
    """
    Check if the public IP is the expected IP. Checks a couple of
    times over max 5 seconds because the route change might not
    take effect immediately. Don't want to give up too soon.
    :param str expected_ip: The expected public IP
    :return str actual_ip: The expected IP if it matches.
    """
    logger.info("Verifying if networking is working as intended")

    try:
        wait_for(
            check_dns,
            value=True,
            check_interval=1,
            timeout=30,
            expect_excs=[RuntimeError],
        )
    except TimeoutException:
        raise RuntimeError("DNS unreachable")
    wait_for(get_public_ip, value=expected_ip, check_interval=1, timeout=5)


def get_manager(
    enable, expected_ip=None, send_slack_message=True, provider=None, interface=None
):
    provider = provider or get_hn_common_var("provider")
    if provider == PROVIDER_DIGITAL_OCEAN:
        return ManageDigitalOceanFloatingIP(enable, expected_ip, send_slack_message)
    elif provider == PROVIDER_TBBM:
        return ManageTBBMFloatingIP(enable, expected_ip, send_slack_message, interface)
    else:
        raise RuntimeError(f"Unexpected provider: {provider}")


def manage_floating_ip(
    enable, expected_ip=None, send_slack_message=True, provider=None, interface=None
):
    manager = get_manager(enable, expected_ip, send_slack_message, provider, interface)
    try:
        manager.validate()
    except SafeRuntimeError as e:
        logger.info(e)
        exit(0)
    try:
        manager.apply()
        sleep(1)
        verify_outbound_ip_is_expected(expected_ip)
        logger.info("Success.")
    except Exception as e:
        manager.revert()
        logger.info("Raising original exception now")
        raise e


def parse_arguments(args=None):
    """
    Parse the hypernode-manage-floating-ip arguments
    :param list args: Optional, defaults to sys.argv
    :return obj args: The parsed arguments
    """
    parser = ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--enable",
        action="store_true",
        help="Enable the use of floating IPs for TBBM or Digital Ocean",
    )
    group.add_argument(
        "--disable",
        action="store_true",
        help="Disable the use of floating IPs for TBBM or Digital Ocean",
    )
    parser.add_argument(
        "--verbose", "-v", help="Display debug output", action="store_true"
    )
    parser.add_argument(
        "--expected-ip",
        help="The expected IP (optional). Will revert if the public IP "
        "is not this IP after performing the networking changes.",
    )
    parser.add_argument(
        "--dev",
        help="Use this flag when testing. Disables slack alerts.",
        action="store_true",
    )
    parser.add_argument(
        "--provider",
        type=str,
        choices=[PROVIDER_DIGITAL_OCEAN, PROVIDER_TBBM],
        help="The provider of this node",
    )
    parser.add_argument(
        "--interface",
        type=str,
        help="Interface to modify. Only necessary on TBBM if the node is unable to reach the internet",
    )
    parsed_args = parser.parse_args(args)
    return parsed_args


def main():
    """
    The entrypoint for this command
    :return None:
    """
    args = parse_arguments()
    configure_logging("kamikaze3.manage_floating_ip", args.verbose)

    # If args.enable is not True, then --disable is set
    manage_floating_ip(
        enable=args.enable,
        expected_ip=args.expected_ip,
        send_slack_message=not args.dev,
        provider=args.provider,
        interface=args.interface,
    )


if __name__ == "__main__":
    main()
