import logging
import shutil
import subprocess
from pathlib import Path
from socket import gethostname

import psutil
import yaml

from kamikaze3.common.env import get_public_ip
from kamikaze3.common.slack import try_send_slack_message
from kamikaze3.manage_floating_ip.settings import (
    NETPLAN_CONFIG_FILE_BACKUP_PATH,
    NETPLAN_CONFIG_FILE_PATH,
)

logger = logging.getLogger(__name__)

NETPLAN_CONFIG_FILE = Path(NETPLAN_CONFIG_FILE_PATH)
NETPLAN_CONFIG_FILE_BACKUP = Path(NETPLAN_CONFIG_FILE_BACKUP_PATH)


# see https://stackoverflow.com/questions/25108581/python-yaml-dump-bad-indentation
# ultimately this does not matter, only if we want to keep the same yaml style
class HypernodeIndentationDumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, False)


class SafeRuntimeError(RuntimeError):
    def __init__(self, message):
        super().__init__(message)


class ManageTBBMFloatingIP:
    _config = None
    config_file_path = NETPLAN_CONFIG_FILE
    config_backup_file_path = NETPLAN_CONFIG_FILE_BACKUP
    subnet = "/32"

    def __init__(
        self, enable, expected_ip=None, send_slack_message=False, interface=None
    ):
        self.enable = enable
        self.expected_ip = expected_ip
        self.send_slack_message = send_slack_message
        if interface:
            self.current_ip = None
            self.interface = interface
        else:
            self.current_ip = get_public_ip()
            self.interface = self._get_interface()

    @property
    def config(self):
        if self._config is None:
            with open(self.config_file_path) as file:
                self._config = yaml.safe_load(file)
        return self._config

    def validate(self):
        if self.enable and not self.expected_ip:
            raise RuntimeError(
                "To enable floating IP you must also supply --expected-ip"
            )
        if self.current_ip == self.expected_ip:
            raise SafeRuntimeError("Public IP is already the expected IP")
        if not self.config_file_path.exists():
            raise RuntimeError(
                f"Cannot manage floating ip: netplan config file not found: {self.config_file_path}"
            )
        self._validate_config()

    def apply(self):
        logger.info("Applying...")
        logger.info(f"Creating a backup, saving it to: {self.config_backup_file_path}")
        shutil.copy(self.config_file_path, self.config_backup_file_path)

        addresses = self.config["network"]["ethernets"][self.interface]["addresses"]
        if self.enable:
            addresses.insert(0, self.expected_ip + self.subnet)
        else:
            del addresses[0]

        with open(self.config_file_path, "w") as file:
            # this would replace the current "### Hypernode installimage"
            file.write("### Edited by hypernode-manage-floating-ip\n")
            yaml.dump(
                self.config,
                file,
                default_flow_style=False,
                sort_keys=False,
                Dumper=HypernodeIndentationDumper,
            )

        # this will cause an exception if it either times out or returns non-zero
        self._apply_config(timeout=10)

    def revert(self):
        logger.info("Reverting...")
        # check backup file
        if not self.config_backup_file_path.exists():
            logger.error(
                f"Cannot revert: backup file not found: {self.config_backup_file_path}"
            )
            return
        logger.info(f"Found config backup: {self.config_backup_file_path}")

        # delete original file
        if self.config_file_path.exists():
            logger.info(f"Deleting file {self.config_file_path}")
            self.config_file_path.unlink()
        else:
            logger.info(f"Could not find {self.config_file_path}. Skipping deleting.")

        # rename backup file
        logger.info(f"Renaming backup file to {self.config_file_path}")
        self.config_backup_file_path.rename(self.config_file_path)

        # netplan apply
        try:
            self._apply_config()
        except subprocess.CalledProcessError as e:
            logger.error(f"Error applying netplan config: {e}")
        if self.send_slack_message:
            try_send_slack_message(
                f":bug: Had to revert hypernode-manage-floating-ip on {gethostname()}"
            )

    def _validate_config(self):
        if not self.enable:
            addresses = self.config["network"]["ethernets"][self.interface]["addresses"]
            if len(addresses) < 2:
                raise SafeRuntimeError(
                    "Won't disable floating IP, it's the only IP in the config"
                )

    def _get_interface(self):
        interfaces = psutil.net_if_addrs()
        for interface, addresses in interfaces.items():
            for addr in addresses:
                if addr.address in (self.current_ip, self.expected_ip):
                    return interface
        raise RuntimeError("Could not find the right interface")

    def _apply_config(self, timeout=None):
        logger.info("Running netplan apply...")
        subprocess.check_call(["/usr/sbin/netplan", "apply"], timeout=timeout)
        logger.info("Successfully applied netplan config")
