import logging
from argparse import ArgumentParser
from collections import defaultdict
from datetime import datetime
from io import StringIO

import boto3
from botocore.exceptions import ClientError
from dateutil.tz import tzlocal
from rich import print
from rich.console import Console

from kamikaze3.common.args import add_default_args
from kamikaze3.common.env import get_hn_object_storage_var
from kamikaze3.common.log import create_syslog_handler
from kamikaze3.object_storage.settings import AWS_DEFAULT_REGION

logger = logging.getLogger(__name__)


def print_and_log(level: int, message: str):
    """
    Prints message to console and logs rendered message to logger
    """
    print(message)

    buffer = StringIO()
    console = Console(file=buffer, force_terminal=True)
    console.print(message)
    logger.log(level, buffer.getvalue().strip())


def get_s3_client() -> boto3.client:
    """
    Get S3 client for connected object storage account
    """
    return boto3.client(
        "s3",
        endpoint_url=get_hn_object_storage_var("endpoint_url"),
        aws_access_key_id=get_hn_object_storage_var("access_key"),
        aws_secret_access_key=get_hn_object_storage_var("secret_key"),
        region_name=AWS_DEFAULT_REGION,
    )


def parse_restore_point(restore_point: str) -> datetime:
    """
    Parse restore point string and return it in local timezone
    """
    return datetime.strptime(restore_point, "%Y-%m-%d %H:%M:%S").astimezone(tzlocal())


def add_arguments_to_parser(parser: ArgumentParser):
    parser.add_argument(
        "bucket_name",
        help="Name of the bucket to restore",
    )
    parser.add_argument(
        "restore_point",
        help="Restore point in time. Format: YYYY-MM-DD HH:MM:SS",
    )
    parser.add_argument(
        "prefix",
        help="Path prefix to restore",
    )
    parser.add_argument(
        "--apply",
        action="store_true",
        help="Actually apply the changes",
        default=False,
    )
    add_default_args(parser)


def parse_args() -> tuple[str, str, str, bool]:
    parser = ArgumentParser(description="Object storage restore tool")
    add_arguments_to_parser(parser)

    args = parser.parse_args()
    return (
        args.bucket_name,
        args.prefix,
        args.restore_point,
        args.apply,
    )


def is_bucket_versioned(s3: boto3.client, bucket_name: str) -> bool:
    """
    Check if the bucket is versioned
    """
    response = s3.get_bucket_versioning(Bucket=bucket_name)
    return response.get("Status", "").lower() == "enabled"


def restore_object_version(
    s3: boto3.client,
    bucket_name: str,
    version: dict,
    delete_markers: list[dict],
    apply: bool = False,
):
    """
    Restore object to given version.
    Before restore happens, DeleteMarkers will first be removed. When removing DeleteMarkers, the object will appear in the bucket again.
    The object appears based on the latest version available in the bucket after the DeleteMarker. If that version appears to be the version
    we want to restore, the restore is considered complete, and we can skip the copy_object call.
    """
    key = version["Key"]
    source_object = {
        "Bucket": bucket_name,
        "Key": key,
        "VersionId": version["VersionId"],
    }

    # If there are DeleteMarkers for this key, delete them first. This is necessary, because the delete markers
    # will prevent the object from being restored.
    for delete_marker in delete_markers:
        if delete_marker["LastModified"] < version["LastModified"]:
            # Skip delete markers that were created before the version
            continue

        if not apply:
            print(f"Would recover deleted object {key}")
            continue

        print_and_log(logging.INFO, f"Recovering deleted object {key}")

        delete_result = s3.delete_object(
            Bucket=bucket_name,
            Key=key,
            VersionId=delete_marker["VersionId"],
        )
        if delete_result.get("VersionId") == version["VersionId"]:
            # The delete marker has been removed and current now points to the target version.
            # This means the object has been fully restored, and we don't need to restore it any further
            logger.debug(f"Deleted object {key} and restored it to target version")
            return

    if not apply:
        print(
            f"Would restore {key} to version {version['LastModified']:%Y-%m-%d %H:%M}"
        )
        return

    print_and_log(
        logging.INFO,
        f"Restoring {key} to version {version['LastModified']:%Y-%m-%d %H:%M}",
    )
    s3.copy_object(
        Bucket=bucket_name,
        Key=key,
        CopySource=source_object,
    )


def restore_bucket(
    s3: boto3.client,
    bucket_name: str,
    prefix: str,
    restore_point: datetime,
    apply: bool,
):
    """
    Restore objects in bucket based on prefix and restore point.
    The restore point will be used as a point in time we want to go back to. The first version that appears before that point in time will be restored.
    """
    if not is_bucket_versioned(s3, bucket_name):
        print_and_log(
            logging.ERROR,
            f"[red]Bucket {bucket_name} is not versioned. Cannot restore.[/]",
        )
        return

    paginator = s3.get_paginator("list_object_versions")
    delete_markers = defaultdict(list)
    restored_objects = set()
    failed_restore_versions = []

    mode = "restore" if apply else "simulation"
    print(f"[bold yellow]Bucket[/]:        {bucket_name}")
    print(f"[bold yellow]Restore point[/]: {restore_point:%Y-%m-%d %H:%M:%S}")
    print(f"[bold yellow]Mode[/]:          {mode}")
    print("")

    # Restore versions from before the restore point
    for page in paginator.paginate(Bucket=bucket_name, Prefix=prefix):
        for delete_marker in page.get("DeleteMarkers", []):
            delete_markers[delete_marker["Key"]].append(delete_marker)
        for version in page.get("Versions", []):
            key = version["Key"]

            if key in restored_objects:
                continue
            if version["LastModified"] > restore_point:
                continue
            if version["IsLatest"]:
                # If this version is the latest, that means the current version for the object is also the most
                # recent version, so we don't need to restore it
                restored_objects.add(key)
                delete_markers.pop(key, None)
                continue

            try:
                restore_object_version(
                    s3,
                    bucket_name,
                    version,
                    list(
                        delete_markers[key]
                    ),  # Make a copy of the list to avoid having mock args reference the same list
                    apply,
                )
                del delete_markers[key]
            except ClientError as e:
                if e.response["Error"]["Code"] == "NoSuchKey":
                    # The object was deleted after the restore point
                    # We will retry restoring this version later
                    print_and_log(
                        logging.WARNING,
                        f"[yellow]Failed to restore, retrying later for {key}:[/] {e}",
                    )
                    failed_restore_versions.append(version)
                else:
                    print_and_log(
                        logging.ERROR, f"[red]Failed to restore object {key}:[/] {e}"
                    )

            restored_objects.add(key)

    # Retry restoring versions that failed to restore.
    # This can happen if the object was deleted after the restore point, and we did not have a DeleteMarker for it yet
    for version in failed_restore_versions:
        key = version["Key"]
        try:
            restore_object_version(s3, bucket_name, version, delete_markers[key], apply)
        except ClientError as e:
            print_and_log(
                logging.ERROR,
                f"[red]Failed second attempt to restore {key}:[/] {e}",
            )
            continue


def main():
    bucket_name, prefix, restore_point, apply = parse_args()
    handle(bucket_name, prefix, restore_point, apply)


def handle(
    bucket_name: str,
    prefix: str,
    restore_point: str,
    apply: bool = False,
    verbose: bool = False,
    **kwargs,
):
    s3 = get_s3_client()
    logger.setLevel(logging.DEBUG)
    logger.addHandler(
        create_syslog_handler(level=logging.DEBUG if verbose else logging.INFO)
    )

    restore_bucket(s3, bucket_name, prefix, parse_restore_point(restore_point), apply)


if __name__ == "__main__":
    main()
