import json
import logging
import subprocess
from datetime import datetime

from kamikaze3.common.env import get_hn_common_var
from kamikaze3.onboarding import try_call_event_hook

ECOMSCAN_BIN = "/usr/local/bin/ecomscan"
ECOMSCAN_RUN_TIMEOUT = 60 * 60 * 8  # 8 Hours
ECOMSCAN_KEY_ENV_VAR = "ECOMSCAN_KEY"

logger = logging.getLogger(__name__)


def run_scan(
    path,
    key=None,
    deep=False,
    skip_dashboard=False,
    skip_database=False,
    skip_symlinks=False,
    enable_notifications=False,
    verbose=False,
):
    """
    Run the Ecomscan command with the provided arguments.
    """
    # If no key is provided, use the default key from Hypernode
    if not key:
        key = get_hn_common_var("ecomscan_key")

    if not key:
        raise RuntimeError("Ecomscan key is required and not provided, aborting.")

    cmd = [
        ECOMSCAN_BIN,
        path,
        "--format json",
    ]
    if deep:
        cmd.append("--deep")
    if skip_dashboard:
        cmd.append("--skip-dashboard")
    if skip_database:
        cmd.append("--skip-database")
    if skip_symlinks:
        cmd.append("--skip-symlinks")
    if verbose:
        cmd.append("--verbose")

    try:
        scan_output = subprocess.check_output(
            " ".join(cmd),
            env={ECOMSCAN_KEY_ENV_VAR: key},
            shell=True,
            timeout=ECOMSCAN_RUN_TIMEOUT,
            encoding="utf-8",
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Ecomscan scan not completed successfully due {e}")

    out = json.loads(scan_output)
    detections = get_malware_detections(out)
    context = out.get("ctx", {})

    scan_date = (
        context["scan_date"]
        if "scan_date" in context
        else datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    )

    if enable_notifications:
        update_api(len(detections), scan_date)


def get_malware_detections(scan_output):
    """
    Scan output and extract the detections of type malware.
    """
    detections = []
    for check in scan_output["check_results"]:
        if not check.get("detections"):
            continue

        for detection in check.get("detections", {}):
            if detection.get("class") != "malware":
                continue

            detections.append(detection)
    return detections


def update_api(vulnerabilities, scan_date):
    event = {
        "malware-scan-finished": {
            "malware_files_found": vulnerabilities,
            "scan_date": scan_date,
        }
    }
    logger.debug(f"Sending malware scan finished event: {event}")
    try_call_event_hook(event)
